"""
Module request monitoring plugin example
"""
from abc import abstractmethod
from logging import Logger
from typing import List, Optional

from aiohttp import ClientSession
from luna_plugins.base.plugins_meta.base_plugins import BaseEventPlugin


class BaseRequestMonitoringPlugin(BaseEventPlugin):
    """
    Base class for requests monitoring.
    """

    # event name for triggering callback
    eventName = "monitoring_event"

    @abstractmethod
    async def flushPointToMonitoring(self, points: List["BaseRequestMonitoringPoint"], logger: Logger) -> None:
        """
        All plugins must realize this method.

        This function call after end of request

        Args:
            points: points for monitoring which corresponding the request
            logger: logger
        """

    async def handleEvent(  # pylint: disable-msg=C0204,W0221
        self, points: List["BaseRequestMonitoringPoint"], logger: Logger
    ):
        await self.flushPointToMonitoring(points=points, logger=logger)


class RequestMonitoringPlugin(BaseRequestMonitoringPlugin):
    """
    Example plugin sends a request data for monitoring to third-party source.
    Only one instance of this class exist during the program execution.
    """

    def __init__(self, app: "LunaApplication"):
        super().__init__(app)
        self.url = "http://127.0.0.1:5020/1/buckets"
        self.session: Optional[ClientSession] = None
        self.bucket = "plugin_test_bucket"

    async def close(self):
        """
        Stop plugin.

        Close all open connections and ect
        """
        if self.session:
            await self.session.close()

    async def initialize(self):
        """
        Initialize plugin.

        Close all open connections and ect
        """
        self.session = ClientSession()
        async with self.session.post(f"{self.url}?bucket={self.bucket}") as resp:
            if resp.status not in (201, 409):
                response = await resp.json()
                raise RuntimeError(f"failed create bucket, {self.bucket}, response: {response}")

    async def flushPointToMonitoring(self, points: List["BaseRequestMonitoringPoint"], logger: Logger) -> None:
        """
        Callback for sending a request monitoring data.

        Args:
            points: point for monitoring which corresponding the request
            logger: logger
        """
        if not points:
            return
        point = points[0]
        logger.debug(
            f"Plugin 'flushPointToMonitoring' get point, request_id: {point.requestId}, "
            f"start time: {point.eventTime}"
        )
        msg = {"tags": point.tags, "fields": point.fields}
        async with self.session.post(f"{self.url}/{self.bucket}/objects", json=msg) as resp:
            logger.info(resp.status)
            logger.info(await resp.text())
