"""
Module with auxiliary redis entities
"""
import abc
import asyncio
import uuid
from asyncio import Task
from dataclasses import dataclass, field
from time import time
from typing import AsyncIterator, Optional

import redis
import ujson
from async_timeout import timeout as async_timeout
from redis.asyncio.client import PubSub
from vlutils.jobs.async_runner import AsyncRunner
from websockets.legacy.protocol import WebSocketCommonProtocol

from ..errors.errors import Error
from ..monitoring.base_monitoring import BaseLunaMonitoring
from ..utils.log import Logger
from .monitoring_points import ErrorPoint

INIT_MASTER_POOL_TIMEOUT = 1


class RedisDisconnected(Exception):
    """Redis disconnected exception"""


class RedisMessage:
    """Container for message from redis

    Attributes:
        rawMsg (str): raw msg from redis
        msg (dict): load json
        receiveTime (float): timestamp of getting message from
    """

    __slots__ = ("rawMsg", "msg", "receiveTime")

    def __init__(self, msg: str) -> None:
        self.receiveTime = time()
        self.rawMsg = msg
        self.msg = ujson.loads(msg["data"])


@dataclass
class BaseWSConnection:
    """
    Base WS connection with filters
    """

    wsResponse: WebSocketCommonProtocol  # a ws connection
    requestId: str  # request id
    subscriptionId: str = field(default_factory=lambda: str(uuid.uuid4()))  # subscription id


class RedisChannel:
    """Abstraction over redis subscribe.

    The chanel assumes subscription on one redis chanel.

    Attributes:
        pubSub: redis-py PubSub
        name: chanel name
    """

    def __init__(self, name: str, pubSub: PubSub):
        self.pubSub = pubSub
        self.name = name

    async def subscribe(self, *args, **kwargs):
        """Activate subscription"""
        return await self.pubSub.subscribe(self.name, *args, **kwargs)

    async def unsubscribe(self, *args):
        """Unsubscribe"""
        return await self.pubSub.unsubscribe(self.name, *args)

    async def listen(self) -> AsyncIterator[str]:
        """Listen messages from chanel"""
        while self.pubSub.subscribed:
            rawResponse = await self.pubSub.parse_response(block=True)
            response = await self.pubSub.handle_message(rawResponse, ignore_subscribe_messages=True)
            if response is not None:
                yield response


class ChannelPublisher(abc.ABC):
    """
    Listen to redis channel and publish events to websockets.

    Attributes:
        channelName (str): channel name for subscription
        redisChannel: A RedisChannel instance - container with queue for messages from channel
        subscribers (Dict[str, WSConnection]): set of current ws subscribers for channel
        wsPublisher: An AsyncRunner instance - manager for scheduling coroutines
        logger (Logger): logger for subscription
        consumerTask: A Task for monitoring messages in redis channel
        monitoring: monitoring client
    """

    def __init__(self, channelName: str, redisChannel: RedisChannel, monitoring: Optional[BaseLunaMonitoring]) -> None:
        self.channelName = channelName
        self.redisChannel = redisChannel
        self.subscribers: dict[str, BaseWSConnection] = {}
        self.logger = Logger(f"subscription {self.channelName}")
        self.wsPublisher = AsyncRunner(1000, exceptionHandler=self.wsPublisherExceptionHandler, closeTimeout=1)
        self.consumerTask: Task | None = None
        self.monitoring = monitoring

    async def consumeRedisChannel(self) -> None:
        """Main loop for consuming messages from redis"""
        self.logger.debug(f"Start consuming redis channel {self.channelName}")
        async for msg in self.redisChannel.listen():
            self.logger.debug(f"Got message from redis {self.channelName}: {msg}")
            await self._processMsgFromRedis(RedisMessage(msg))

    @property
    def hasSubsribers(self) -> bool:
        """Returns whether there are active subscribers."""
        return bool(self.subscribers)

    def wsPublisherExceptionHandler(self, _: AsyncRunner, context: dict) -> None:
        """
        Exception callback for aiojobs.Scheduler

        Fails task.

        Args:
            _: scheduler with requests jobs (non used, for interface)
            context: context is a dict with the following keys:
                     - message: error message, str
                     - job: failed job, Job instance
                     - exception: caught exception, Exception instance

        """
        error = Error.FailToPublishMessageToWS
        self.logger.error(f"{error.asDict()}, message: {context['message']}")
        if self.monitoring:
            point = ErrorPoint(time(), errorCode=error.errorCode, additionalTags={"channel": self.channelName})
            self.monitoring.flushPoints([point])

    async def closePublisher(self, timeout: float | None) -> None:
        """
        Cleanup
        """
        await self.unsubscribeFromRedisChannel(timeout)
        if self.wsPublisher is not None:
            await self.wsPublisher.close()
            self.wsPublisher = None

        if self.consumerTask is not None:
            self.consumerTask.cancel()
            self.consumerTask = None

    def __setitem__(self, subscriptionId: str, wsConnection: BaseWSConnection):
        """
        Add new ws connection to subscribers.

        Args:
            subscriptionId: subscription id
            wsConnection: ws connection
        """
        self.subscribers[subscriptionId] = wsConnection

    @abc.abstractmethod
    def getSubscribersForEvent(self, *args, **kwargs) -> list[BaseWSConnection]:
        """
        Get subscribers for input event.

        Returns:
            list of subscribers
        """

    @abc.abstractmethod
    async def processMsgFromRedis(self, redisMsg: RedisMessage) -> None:
        """
        Process msg from redis and send events to ws

        Args:
            redisMsg: msg from redis
        """

    async def _processMsgFromRedis(self, redisMsg: RedisMessage) -> None:
        """
        Wrap for process msg from redis and send events to ws

        Args:
            redisMsg: msg from redis
        """
        try:
            await self.processMsgFromRedis(redisMsg)
        except Exception:  # pylint: disable-msg=broad-except
            self.logger.exception()
            error = Error.FailToProcessMessage
            if self.monitoring:
                point = ErrorPoint(time(), error.errorCode, additionalTags={"channel": self.channelName})
                self.monitoring.flushPoints([point])

    def __contains__(self, subscriptionId: str) -> bool:
        """Returns whether we have subscription with such id."""
        return subscriptionId in self.subscribers

    def __delitem__(self, subscriptionId: str):
        """
        Remove ws subscription.

        Args:
            subscriptionId: subscription id
        """
        del self.subscribers[subscriptionId]

    async def unsubscribeFromRedisChannel(self, timeout: float | None = None) -> None:
        """
        Wrapper around redis.unsubscribe().

        Args:
            timeout: optional timeout

        In case of error or timeout simply returns.
        """
        try:
            async with async_timeout(timeout):
                await self.redisChannel.unsubscribe()
        except (redis.exceptions.ConnectionError, asyncio.TimeoutError) as exc:
            self.logger.warning(f"Unsubscribe from {self.redisChannel.name} failed: {exc}")

    async def subscribeToRedisChannel(self, subscribeAttempts: int = 1, timeout: float | None = None) -> None:
        """Wrapper around redis.subscribe().

        Args:
            timeout: optional timeout
            subscribeAttempts: subscription attempt count
        Raises:
            RedisDisconnected: If subscribe failed for some reason
        """

        for attemptNumber in range(subscribeAttempts):
            try:
                async with async_timeout(timeout):
                    await self.redisChannel.subscribe()
                self.consumerTask = asyncio.create_task(self.consumeRedisChannel())
                return
            except (redis.exceptions.ConnectionError, asyncio.TimeoutError):
                self.logger.error(f"Attempt #{attemptNumber+1} to subscribe to {self.redisChannel.name} failed...")

            await asyncio.sleep(attemptNumber * 1)

        raise RedisDisconnected()

    async def reconnect(self):
        """Reconnect to redis chanel"""
        await self.redisChannel.unsubscribe()
        self.consumerTask.cancel()
        await self.redisChannel.subscribe()
        self.consumerTask = asyncio.create_task(self.consumeRedisChannel())
