"""
This module defines classes and functions for encoding/decoding python objects
to/from msgpack.

Integer constants are used to help deserializer select the right type. Maximum value is 127, minimum is 1.
see msgpack ExtType https://github.com/msgpack/msgpack/blob/master/spec.md#ext-format-family
"""
from datetime import datetime
from functools import partial
from typing import Any, Callable

import msgpack

EXT_CODE_TEMPORARY_ATTRIBUTES = 1
EXT_CODE_BASIC_ATTRIBUTES = 2
EXT_CODE_DESCRIPTOR = 3
EXT_CODE_DATETIME = 127


class Registry:
    """
    Registry for serializers and deserializers

    Attributes:
        serializers: A dict, mapping between type and appropriate serializer
        deserializers: A dict, mapping between code and deserializer
    """

    def __init__(self) -> None:
        """Initialize state"""
        self.serializers = {}
        self.deserializers = {}

    def __contains__(self, code: int) -> bool:
        """Returns whether code already registered"""
        return code in self.deserializers

    def register(self, extCode: int, objType: Any, serializeFunc: Callable, deserializeFunc: Callable) -> None:
        """
        Register type

        Args:
            extCode: An integer, unique type code
            objType: A type of object
            serializeFunc: A function for objects serialization
            deserializeFunc: A function for objects deserialization
        """
        self.serializers[objType] = serializeFunc
        self.deserializers[extCode] = deserializeFunc

    def serialize(self, obj: Any) -> msgpack.ExtType:
        """
        Serialize object with appropriate serializer

        Args:
            obj: Any object

        Raises:
            TypeError if object is not registered

        Returns:
            msgpack.ExtType with encoded object
        """
        if type(obj) not in self.serializers:  # pylint: disable=unidiomatic-typecheck
            raise TypeError(f"Attempt to serialize unregistered type '{type(obj)}'")

        serializerFunc = self.serializers[type(obj)]
        return serializerFunc(obj)

    def deserialize(self, extCode: int, data: bytes) -> Any:
        """
        Deserialize object with appropriate serializer

        Args:
            data: A bytes with object state
            extCode: An integer, unique type code

        Raises:
            TypeError if type code is not registered

        Returns:
            Approripate object instance
        """
        if extCode not in self.deserializers:
            raise TypeError(f"Attempt to deserialize unregistered type '{extCode}'")

        deserializeFunc = self.deserializers[extCode]
        return deserializeFunc(data)


def dumpb(instance: Any) -> bytes:
    """Wrapper around msgpack.packb

    Allows recursive object serialization

    Args:
        instance: Any object

    Returns:
        a bytes, serialized object
    """
    return msgpack.packb(instance, default=registry.serialize, use_bin_type=True)


def loadb(data: bytes) -> Any:
    """Wrapper around msgpack.unpackb

    Allows recursive object deserialization

    Args:
        data: A bytes, serialized object

    Returns:
        any object
    """
    return msgpack.unpackb(data, ext_hook=registry.deserialize, raw=False)


def _serializeDatetime(obj: datetime) -> msgpack.ExtType:
    """Serializer for datetime type

    Args:
        obj: A datetime instance

    Returns:
        msgpack.ExtType with encoded object
    """
    return msgpack.ExtType(EXT_CODE_DATETIME, dumpb(obj.isoformat()))


def _deserializeDatetime(data: bytes) -> datetime:
    """Deserializer for datetime type

    Args:
        data: bytes

    Returns:
        datetime instance
    """
    return datetime.fromisoformat(loadb(data))


def _serializeModel(extCode: int, obj: Any) -> msgpack.ExtType:
    """Serialize obj

    Args:
        extCode: A int from 1 to 126, unique type identifier
        obj: An instance of object to serialize

    Returns:
        msgpack.ExtType with encoded object
    """
    state = obj.__getstate__()
    return msgpack.ExtType(extCode, dumpb(state))


def _deserializeModel(cls, data: bytes) -> Any:
    """Deserialize obj from bytes.

    Args:
        cls: A Type of object
        data: A bytes with object state

    Returns:
        Object with restored state
    """
    state = loadb(data)
    instance = cls.__new__(cls)
    instance.__setstate__(state)
    return instance


class SlotsStateMixin:
    """Generic protocol realization for models"""

    def __setstate__(self, state: dict) -> None:
        """
        Set object state

        Args:
            state: A dictionary with state
        """
        for key, value in state.items():
            setattr(self, key, value)

    def __getstate__(self) -> dict:
        """
        Get object state

        Returns:
            A dict with state
        """
        state = {slot: getattr(self, slot) for slot in self.__slots__}
        return state


def registerModel(extCode: int) -> Callable:
    """
    Decorator fabric

    Args:
        extCode: An integer, unique type code

    Raises:
        TypeError if model doesnt realize required protocol or extCode is not uniq

    Returns:
        decorator for model registration
    """

    def decorator(cls):
        if not hasattr(cls, "__getstate__"):
            raise TypeError(f"{cls.__name__} must implement __getstate__")

        if not hasattr(cls, "__setstate__"):
            raise TypeError(f"{cls.__name__} must implement __setstate__")

        if extCode in registry:
            raise TypeError(f"{extCode} already registered")

        registry.register(extCode, cls, partial(_serializeModel, extCode), partial(_deserializeModel, cls))

        return cls

    return decorator


registry = Registry()  # pylint: disable=invalid-name
registry.register(EXT_CODE_DATETIME, datetime, _serializeDatetime, _deserializeDatetime)
