"""
Module realizes service error handler
"""
from asyncio import CancelledError
from typing import Optional, Tuple, Union

from httptools.parser.errors import HttpParserError  # type: ignore
from luna3.common.exceptions import LunaApiException
from luna3.common.luna_response import LunaResponse
from luna_plugins.utils.exceptions import LunaPluginException
from sanic.exceptions import (
    InvalidUsage,
    MethodNotSupported,
    NotFound,
    PayloadTooLarge,
    RequestTimeout,
    SanicException,
    ServiceUnavailable,
)
from sanic.response import HTTPResponse

from ..errors.errors import Error, ErrorInfo
from ..errors.exception import VLException
from ..utils.log import Logger
from ..utils.rid import generateRequestId
from ..web import response
from ..web.request import VLRequest


def createErrorFromLunaApiException(exception: Union[LunaApiException, LunaResponse]) -> ErrorInfo:
    """
    Generate from LunaApiException a ErrorInfo
    Args:
        exception: exception from luna3

    Returns:
        ErrorInfo
    """
    if "error" in exception.json:
        error = ErrorInfo(exception.json["error"], exception.json["desc"], exception.json["detail"])
    else:
        error = ErrorInfo.fromDict(exception.json)
    return error


def _handleSanicException(request: VLRequest, exception: SanicException) -> Tuple[int, ErrorInfo]:
    """
    Convert sanic exception to error and status code
    Args:
        request: request
        exception: exception

    Returns:
        <status code, error>
    """
    errorStatusCode = exception.status_code
    if isinstance(exception, ServiceUnavailable):
        if str(exception) == "Response Timeout":
            error = Error.ResponseTimeout.format(request.app.config["RESPONSE_TIMEOUT"])
            errorStatusCode = 503  # sanic status code for this case
        elif str(exception) == "Cancelled":
            error = Error.ClientDisconected
            errorStatusCode = 499
        else:
            error = Error.UnknownWebApplicationError.format(exception)
    elif isinstance(exception, RequestTimeout):
        error = Error.RequestTimeout.format(request.app.config["REQUEST_TIMEOUT"])
        errorStatusCode = 408  # rest status code for this case
    elif isinstance(exception, PayloadTooLarge):
        error = Error.PayloadToLarge
        errorStatusCode = 413  # rest status code for this case
    else:
        error = Error.UnknownWebApplicationError.format(exception)
    return errorStatusCode, error


def serverErrorHandler(request: Optional[VLRequest], exception: Exception) -> HTTPResponse:
    """
    Error handler.

    Handle exception ang generate  correct response with error
    Args:
        request: request
        exception: exception

    Returns:
        response with error
    """
    # pylint: disable-msg=too-many-branches
    if request is not None:
        logger = request.logger
    else:
        # Happens in case of HttpParserError
        requestId = generateRequestId(Logger.logTime == "LOCAL")  # pylint: disable-msg=comparison-with-callable
        logger = Logger(requestId)

    errorStatusCode = 500
    errorHeaders = None
    if isinstance(exception, VLException):
        if exception.isCriticalError:
            logger.exception()
        error = exception.error
        errorStatusCode = exception.statusCode
        errorHeaders = exception.headers
    elif isinstance(exception, LunaApiException):
        if exception.request.method == "HEAD":
            error = None
        else:
            error = createErrorFromLunaApiException(exception)
        errorStatusCode = exception.statusCode
    elif isinstance(exception, NotFound):
        errorStatusCode = 404
        error = Error.PageNotFoundError
    elif isinstance(exception, MethodNotSupported):
        error = Error.MethodNotAllowed
        errorStatusCode = 405
    elif isinstance(exception, CancelledError):
        errorStatusCode = 499
        error = Error.UnknownError
    elif isinstance(exception, HttpParserError):
        errorStatusCode = 400
        error = Error.InvalidHttpRequest.format(exception)
    elif isinstance(exception, InvalidUsage):
        errorStatusCode = exception.status_code
        error = Error.InvalidRequest.format(str(exception))
    elif isinstance(exception, LunaPluginException):
        error = Error.generateError(Error.UnknownError, str(exception))
    elif isinstance(exception, SanicException):
        errorStatusCode, error = _handleSanicException(request, exception)
    else:
        logger.exception()
        error = Error.RequestInternalServerError
    jsonResp = response.json(body=error.asDict(), status=errorStatusCode, headers=errorHeaders)
    # set error info to response (required for middleware)
    request.error = error
    return jsonResp
