"""
Rest validators
"""
from typing import Literal

from vlutils.descriptors.data import DescriptorsEnum

from ..errors.errors import Error, ErrorInfo
from ..errors.exception import VLException

META_OPERATOR = Literal["like", "nlike", "eq", "neq", "in", "nin", "gt", "gte", "lt", "lte"]


def validateDescriptor(descriptor: bytes, version: int, raiseError: bool = True) -> None | ErrorInfo:
    """
    Validate descriptor
    Args:
        descriptor: raw descriptor
        version: version
        raiseError: raise error or return error
    Returns:
        None if descriptor is valid otherwise error
    Raises:
        VLException: if descriptor is not valid and raiseError is True
    """
    error = None
    for descriptorType in DescriptorsEnum:
        if descriptorType.value.version == version:
            if descriptorType.value.length != len(descriptor):
                error = Error.InvalidDescriptorLength.format(len(descriptor))
            break
    else:
        error = Error.UnknownDescriptorVersion.format(version)
    if error is not None and raiseError:
        raise VLException(isCriticalError=False, statusCode=400, error=error)
    return error


def validateMetaFilter(value: object, op: META_OPERATOR):
    """
    Args:
        value: value to validate
        op: name of the operator
    Raises:
        ValueError in case specified combination of value and operator is invalid
    """

    if isinstance(value, dict):
        raise ValueError("Object in value is not allowed")

    if isinstance(value, bool):
        raise ValueError("Boolean is not allowed")

    if op in ("like", "nlike") and not isinstance(value, str):
        raise ValueError("Pattern should be a string")

    if op in ("in", "nin"):
        if not isinstance(value, list):
            raise ValueError("Value should be a list")

        # [1, 2, "3"] - invalid
        # [1, 2, None] - valid
        allowed = {int, str, float}
        seen = set()
        for x in value:
            if x is None:
                continue

            seen.add(type(x))
            if len(seen) > 1:
                raise ValueError("Only one type allowed")

        if seen - allowed:
            # list contains unallowed types
            raise ValueError("Not allowed type")

        if not seen:
            # list of empty items or [null, null]
            raise ValueError("Empty list or can't deterine datatype to use")

    elif isinstance(value, list):
        # value eq: [] not allowed
        raise ValueError("This combination of value and operator is not allowed")

    if value is None and op not in ("eq", "neq"):
        raise ValueError("Null is allowed with eq/neq")
